@@ -5,13 +5,17 @@
 
 from __future__ import unicode_literals
 
+import sys
 import socket
-from unittest import TestCase
 
+from .common import HyperlinkTestCase
 from .. import URL, URLParseError
 # automatically import the py27 windows implementation when appropriate
-from .._url import inet_pton, SCHEME_PORT_MAP
+from .. import _url
+from .._url import inet_pton, SCHEME_PORT_MAP, parse_host
 
+
+PY2 = (sys.version_info[0] == 2)
 unicode = type(u'')
 
 
@@ -115,10 +119,33 @@ ROUNDTRIP_TESTS = (
      '20Linux%20precise-5.7.1.iso&tr=udp://tracker.openbittorrent.com:80&'
      'tr=udp://tracker.publicbt.com:80&tr=udp://tracker.istole.it:6969&'
      'tr=udp://tracker.ccc.de:80&tr=udp://open.demonii.com:1337'),
+
+    # percent-encoded delimiters in percent-encodable fields
+
+    'https://%3A@example.com/',  # colon in username
+    'https://%40@example.com/',  # at sign in username
+    'https://%2f@example.com/',  # slash in username
+    'https://a:%3a@example.com/',  # colon in password
+    'https://a:%40@example.com/',  # at sign in password
+    'https://a:%2f@example.com/',  # slash in password
+    'https://a:%3f@example.com/',  # question mark in password
+    'https://example.com/%2F/',  # slash in path
+    'https://example.com/%3F/',  # question mark in path
+    'https://example.com/%23/',  # hash in path
+    'https://example.com/?%23=b',  # hash in query param name
+    'https://example.com/?%3D=b',  # equals in query param name
+    'https://example.com/?%26=b',  # ampersand in query param name
+    'https://example.com/?a=%23',  # hash in query param value
+    'https://example.com/?a=%26',  # ampersand in query param value
+    'https://example.com/?a=%3D',  # equals in query param value
+    # double-encoded percent sign in all percent-encodable positions:
+    "http://(%2525):(%2525)@example.com/(%2525)/?(%2525)=(%2525)#(%2525)",
+    # colon in first part of schemeless relative url
+    'first_seg_rel_path__colon%3Anotok/second_seg__colon%3Aok',
 )
 
 
-class TestURL(TestCase):
+class TestURL(HyperlinkTestCase):
     """
     Tests for L{URL}.
     """
@@ -231,9 +258,21 @@ class TestURL(TestCase):
         L{URL.to_text} should invert L{URL.from_text}.
         """
         for test in ROUNDTRIP_TESTS:
-            result = URL.from_text(test).to_text()
+            result = URL.from_text(test).to_text(with_password=True)
             self.assertEqual(test, result)
 
+    def test_roundtrip_double_iri(self):
+        for test in ROUNDTRIP_TESTS:
+            url = URL.from_text(test)
+            iri = url.to_iri()
+            double_iri = iri.to_iri()
+            assert iri == double_iri
+
+            iri_text = iri.to_text(with_password=True)
+            double_iri_text = double_iri.to_text(with_password=True)
+            assert iri_text == double_iri_text
+        return
+
     def test_equality(self):
         """
         Two URLs decoded using L{URL.from_text} will be equal (C{==}) if they
@@ -245,7 +284,7 @@ class TestURL(TestCase):
         self.assertNotEqual(
             urlpath,
             URL.from_text('ftp://www.anotherinvaliddomain.com/'
-                         'foo/bar/baz/?zot=21&zut')
+                          'foo/bar/baz/?zot=21&zut')
         )
 
     def test_fragmentEquality(self):
@@ -294,6 +333,13 @@ class TestURL(TestCase):
         self.assertTrue(childURL.rooted)
         self.assertEqual("http://www.foo.com/c", childURL.to_text())
 
+    def test_emptyChild(self):
+        """
+        L{URL.child} without any new segments returns the original L{URL}.
+        """
+        url = URL(host=u"www.foo.com")
+        self.assertEqual(url.child(), url)
+
     def test_sibling(self):
         """
         L{URL.sibling} of a L{URL} replaces the last path segment, but does not
@@ -352,10 +398,15 @@ class TestURL(TestCase):
         res = 'http://hatnote.com/a/c/d/'
         self.assertEqual(u.click('').to_text(), res)
 
-
         # test click default arg is same as empty string above
         self.assertEqual(u.click().to_text(), res)
 
+        # test click on a URL instance
+        u = URL.fromText('http://localhost/foo/?abc=def')
+        u2 = URL.from_text('bar')
+        u3 = u.click(u2)
+        self.assertEqual(u3.to_text(), 'http://localhost/foo/bar')
+
     def test_clickRFC3986(self):
         """
         L{URL.click} should correctly resolve the examples in RFC 3986.
@@ -476,7 +527,7 @@ class TestURL(TestCase):
 
     def test_queryRemove(self):
         """
-        L{URL.remove} removes all instances of a query parameter.
+        L{URL.remove} removes instances of a query parameter.
         """
         url = URL.from_text(u"https://example.com/a/b/?foo=1&bar=2&foo=3")
         self.assertEqual(
@@ -484,6 +535,21 @@ class TestURL(TestCase):
             URL.from_text(u"https://example.com/a/b/?bar=2")
         )
 
+        self.assertEqual(
+            url.remove(name=u"foo", value=u"1"),
+            URL.from_text(u"https://example.com/a/b/?bar=2&foo=3")
+        )
+
+        self.assertEqual(
+            url.remove(name=u"foo", limit=1),
+            URL.from_text(u"https://example.com/a/b/?bar=2&foo=3")
+        )
+
+        self.assertEqual(
+            url.remove(name=u"foo", value=u"1", limit=0),
+            URL.from_text(u"https://example.com/a/b/?foo=1&bar=2&foo=3")
+        )
+
     def test_parseEqualSignInParamValue(self):
         """
         Every C{=}-sign after the first in a query parameter is simply included
@@ -491,10 +557,16 @@ class TestURL(TestCase):
         """
         u = URL.from_text('http://localhost/?=x=x=x')
         self.assertEqual(u.get(''), ['x=x=x'])
-        self.assertEqual(u.to_text(), 'http://localhost/?=x%3Dx%3Dx')
+        self.assertEqual(u.to_text(), 'http://localhost/?=x=x=x')
         u = URL.from_text('http://localhost/?foo=x=x=x&bar=y')
         self.assertEqual(u.query, (('foo', 'x=x=x'), ('bar', 'y')))
-        self.assertEqual(u.to_text(), 'http://localhost/?foo=x%3Dx%3Dx&bar=y')
+        self.assertEqual(u.to_text(), 'http://localhost/?foo=x=x=x&bar=y')
+
+        u = URL.from_text('https://example.com/?argument=3&argument=4&operator=%3D')
+        iri = u.to_iri()
+        self.assertEqual(iri.get('operator'), ['='])
+        # assert that the equals is not unnecessarily escaped
+        self.assertEqual(iri.to_uri().get('operator'), ['='])
 
     def test_empty(self):
         """
@@ -694,10 +766,14 @@ class TestURL(TestCase):
     def test_queryIterable(self):
         """
         When a L{URL} is created with a C{query} argument, the C{query}
-        argument is converted into an N-tuple of 2-tuples.
+        argument is converted into an N-tuple of 2-tuples, sensibly
+        handling dictionaries.
         """
+        expected = (('alpha', 'beta'),)
         url = URL(query=[['alpha', 'beta']])
-        self.assertEqual(url.query, (('alpha', 'beta'),))
+        self.assertEqual(url.query, expected)
+        url = URL(query={'alpha': 'beta'})
+        self.assertEqual(url.query, expected)
 
     def test_pathIterable(self):
         """
@@ -724,7 +800,7 @@ class TestURL(TestCase):
             def __repr__(self):
                 return "<unexpected>"
 
-        defaultExpectation = "unicode"
+        defaultExpectation = "unicode" if bytes is str else "str"
 
         def assertRaised(raised, expectation, name):
             self.assertEqual(str(raised.exception),
@@ -821,9 +897,37 @@ class TestURL(TestCase):
         url = URL.from_text(t)
         assert url.host == '2001:0db8:85a3:0000:0000:8a2e:0370:7334'
         assert url.port == 80
-        assert url.family == socket.AF_INET6
         assert SCHEME_PORT_MAP[url.scheme] != url.port
 
+    def test_basic(self):
+        text = 'https://user:pass@example.com/path/to/here?k=v#nice'
+        url = URL.from_text(text)
+        assert url.scheme == 'https'
+        assert url.userinfo == 'user:pass'
+        assert url.host == 'example.com'
+        assert url.path == ('path', 'to', 'here')
+        assert url.fragment == 'nice'
+
+        text = 'https://user:pass@127.0.0.1/path/to/here?k=v#nice'
+        url = URL.from_text(text)
+        assert url.scheme == 'https'
+        assert url.userinfo == 'user:pass'
+        assert url.host == '127.0.0.1'
+        assert url.path == ('path', 'to', 'here')
+
+        text = 'https://user:pass@[::1]/path/to/here?k=v#nice'
+        url = URL.from_text(text)
+        assert url.scheme == 'https'
+        assert url.userinfo == 'user:pass'
+        assert url.host == '::1'
+        assert url.path == ('path', 'to', 'here')
+
+    def test_invalid_url(self):
+        self.assertRaises(URLParseError, URL.from_text, '#\n\n')
+
+    def test_invalid_authority_url(self):
+        self.assertRaises(URLParseError, URL.from_text, 'http://abc:\n\n/#')
+
     def test_invalid_ipv6(self):
         invalid_ipv6_ips = ['2001::0234:C1ab::A0:aabc:003F',
                             '2001::1::3F',
@@ -836,16 +940,6 @@ class TestURL(TestCase):
                               socket.AF_INET6, ip)
             self.assertRaises(URLParseError, URL.from_text, url_text)
 
-    def test_ip_family_detection(self):
-        u = URL.from_text('http://giggle.com')
-        self.assertEqual(u.family, None)
-
-        u = URL.from_text('http://127.0.0.1/a/b/?c=d')
-        self.assertEqual(u.family, socket.AF_INET)
-
-        u = URL.from_text('http://[::1]/a/b/?c=d')
-        self.assertEqual(u.family, socket.AF_INET6)
-
     def test_invalid_port(self):
         self.assertRaises(URLParseError, URL.from_text, 'ftp://portmouth:smash')
         self.assertRaises(ValueError, URL.from_text,
@@ -906,6 +1000,13 @@ class TestURL(TestCase):
         url2 = url.replace(scheme='mailto')
         self.assertEquals(url2.to_text(), 'mailto:/path/to/heck')
 
+        url_text = 'unregisteredscheme:///a/b/c'
+        url = URL.from_text(url_text)
+        no_netloc_url = url.replace(uses_netloc=False)
+        self.assertEquals(no_netloc_url.to_text(), 'unregisteredscheme:/a/b/c')
+        netloc_url = url.replace(uses_netloc=True)
+        self.assertEquals(netloc_url.to_text(), url_text)
+
         return
 
     def test_wrong_constructor(self):
@@ -1021,68 +1122,89 @@ class TestURL(TestCase):
 
         assert u1 == u2
 
+    def test_from_text_type(self):
+        assert URL.from_text(u'#ok').fragment == u'ok'  # sanity
+        self.assertRaises(TypeError, URL.from_text, b'bytes://x.y.z')
+        self.assertRaises(TypeError, URL.from_text, object())
 
-    # python 2.6 compat
-    def assertRaises(self, excClass, callableObj=None, *args, **kwargs):
-        """Fail unless an exception of class excClass is raised
-           by callableObj when invoked with arguments args and keyword
-           arguments kwargs. If a different type of exception is
-           raised, it will not be caught, and the test case will be
-           deemed to have suffered an error, exactly as for an
-           unexpected exception.
+    def test_from_text_bad_authority(self):
+        # bad ipv6 brackets
+        self.assertRaises(URLParseError, URL.from_text, 'http://[::1/')
+        self.assertRaises(URLParseError, URL.from_text, 'http://::1]/')
+        self.assertRaises(URLParseError, URL.from_text, 'http://[[::1]/')
+        self.assertRaises(URLParseError, URL.from_text, 'http://[::1]]/')
 
-           If called with callableObj omitted or None, will return a
-           context object used like this::
+        # empty port
+        self.assertRaises(URLParseError, URL.from_text, 'http://127.0.0.1:')
+        # non-integer port
+        self.assertRaises(URLParseError, URL.from_text, 'http://127.0.0.1:hi')
+        # extra port colon (makes for an invalid host)
+        self.assertRaises(URLParseError, URL.from_text, 'http://127.0.0.1::80')
 
-                with self.assertRaises(SomeException):
-                    do_something()
+    def test_normalize(self):
+        url = URL.from_text('HTTP://Example.com/A%61/./../A%61?B%62=C%63#D%64')
+        assert url.get('Bb') == []
+        assert url.get('B%62') == ['C%63']
+        assert len(url.path) == 4
 
-           The context manager keeps a reference to the exception as
-           the 'exception' attribute. This allows you to inspect the
-           exception after the assertion::
+        # test that most expected normalizations happen
+        norm_url = url.normalize()
 
-               with self.assertRaises(SomeException) as cm:
-                   do_something()
-               the_exception = cm.exception
-               self.assertEqual(the_exception.error_code, 3)
-        """
-        context = _AssertRaisesContext(excClass, self)
-        if callableObj is None:
-            return context
-        with context:
-            callableObj(*args, **kwargs)
+        assert norm_url.scheme == 'http'
+        assert norm_url.host == 'example.com'
+        assert norm_url.path == ('Aa',)
+        assert norm_url.get('Bb') == ['Cc']
+        assert norm_url.fragment == 'Dd'
+        assert norm_url.to_text() == 'http://example.com/Aa?Bb=Cc#Dd'
 
+        # test that flags work
+        noop_norm_url = url.normalize(scheme=False, host=False,
+                                      path=False, query=False, fragment=False)
+        assert noop_norm_url == url
 
-# PYTHON 2.6 compat
+        # test that empty paths get at least one slash
+        slashless_url = URL.from_text('http://example.io')
+        slashful_url = slashless_url.normalize()
+        assert slashful_url.to_text() == 'http://example.io/'
 
-class _AssertRaisesContext(object):
-    """A context manager used to implement TestCase.assertRaises* methods."""
+        # test case normalization for percent encoding
+        delimited_url = URL.from_text('/a%2fb/cd%3f?k%3d=v%23#test')
+        norm_delimited_url = delimited_url.normalize()
+        assert norm_delimited_url.to_text() == '/a%2Fb/cd%3F?k%3D=v%23#test'
 
-    def __init__(self, expected, test_case, expected_regexp=None):
-        self.expected = expected
-        self.failureException = test_case.failureException
-        self.expected_regexp = expected_regexp
+        # test invalid percent encoding during normalize
+        assert URL(path=('', '%te%sts')).normalize(percents=False).to_text() == '/%te%sts'
+        assert URL(path=('', '%te%sts')).normalize().to_text() == '/%25te%25sts'
 
-    def __enter__(self):
-        return self
+        percenty_url = URL(scheme='ftp', path=['%%%', '%a%b'], query=[('%', '%%')], fragment='%', userinfo='%:%')
 
-    def __exit__(self, exc_type, exc_value, tb):
-        if exc_type is None:
-            try:
-                exc_name = self.expected.__name__
-            except AttributeError:
-                exc_name = str(self.expected)
-            raise self.failureException(
-                "{0} not raised".format(exc_name))
-        if not issubclass(exc_type, self.expected):
-            # let unexpected exceptions pass through
-            return False
-        self.exception = exc_value # store for later retrieval
-        if self.expected_regexp is None:
-            return True
+        assert percenty_url.to_text(with_password=True) == 'ftp://%:%@/%%%/%a%b?%=%%#%'
+        assert percenty_url.normalize().to_text(with_password=True) == 'ftp://%25:%25@/%25%25%25/%25a%25b?%25=%25%25#%25'
 
-        expected_regexp = self.expected_regexp
-        if not expected_regexp.search(str(exc_value)):
-            raise self.failureException('"%s" does not match "%s"' %
-                     (expected_regexp.pattern, str(exc_value)))
-        return True
+    def test_str(self):
+        # see also issue #49
+        text = u'http://example.com/á/y%20a%20y/?b=%25'
+        url = URL.from_text(text)
+        assert unicode(url) == text
+        assert bytes(url) == b'http://example.com/%C3%A1/y%20a%20y/?b=%25'
+
+        if PY2:
+            assert isinstance(str(url), bytes)
+            assert isinstance(unicode(url), unicode)
+        else:
+            assert isinstance(str(url), unicode)
+            assert isinstance(bytes(url), bytes)
+
+    def test_idna_corners(self):
+        text = u'http://abé.com/'
+        url = URL.from_text(text)
+        assert url.to_iri().host == u'abé.com'
+        assert url.to_uri().host == u'xn--ab-cja.com'
+
+        url = URL.from_text("http://ドメイン.テスト.co.jp#test")
+        assert url.to_iri().host == u'ドメイン.テスト.co.jp'
+        assert url.to_uri().host == u'xn--eckwd4c7c.xn--zckzah.co.jp'
+
+        assert url.to_uri().get_decoded_url().host == u'ドメイン.テスト.co.jp'
+
+        assert URL.from_text('http://Example.com').to_uri().get_decoded_url().host == 'example.com'
